#
#	$Endicor: TinyTable.py,v 1.23 1999/01/10 23:09:51 tsarna Exp tsarna $
#
# Copyright (c) 1998-1999 Endicor Technologies, Inc.
# All rights reserved. Written by Ty Sarna <tsarna@endicor.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
__doc__='''Tiny Table data manager product

$Endicor: TinyTable.py,v 1.23 1999/01/10 23:09:51 tsarna Exp tsarna $'''

__version__='$Revision: 1.23 $'[11:-2]

######################### Imported Modules #########################

from Globals import HTMLFile, MessageDialog, Persistent
from Shared.DC.ZRDB.Results import Results
from PersistentMapping import PersistentMapping
from DateTime import DateTime
from App.Extensions import getBrain
import OFS.ObjectManager, OFS.SimpleItem, Acquisition, AccessControl.Role
import Record, string, types

######################### Folder Methods #########################

addItemForm=HTMLFile('Add', globals())
def addItem(self, id, title='', delim_char='', columns='', REQUEST=None):
    """Add a TinyTable to a folder

    The argument 'self' will be bound to a folder.

    The arguments, 'id' and 'title' are minimal, and most applications
    will have additional needed arguments.
    """
    self._setObject(id, TinyTable(id, title, delim_char, columns))
    if REQUEST is not None: return self.manage_main(self,REQUEST)

######################### Helper Functions #########################

IntValued = 'i'
LongValued = 'l'
FloatValued = 'n'
StringValued = 's'
DateValued = 'd'

TypeNames = {
    IntValued : ':int',
    LongValued : ':long',
    FloatValued : ':float',
    StringValued : '',
    DateValued : ':date'
}

def TypeCode(t):
    if t == LongValued:
        return 'i'
    else:
        return t

def CoerceType(x, t):
    if t == IntValued:
        if type(x) != type(0):
            if (type(x) != type('')):
                x = str(x)

            try:
                x = string.atoi(x)
            except:
                x = 0
        return x
    elif t == LongValued:
        if type(x) != type(0L):
            if (type(x) != type('')):
                x = str(x)

            try:
                x = string.atol(x)
            except:
                x = 0L
        return x
    elif t == FloatValued:
        if type(x) != type(0.0):
            if (type(x) != type('')):
                x = str(x)

            try:
                x = string.atof(x)
            except:
                x = 0.0
        return x
    elif t == DateValued:
        if not (type(x) == types.InstanceType and x.__class__ == DateTime):
            try:
                x = DateTime(x)
            except:
                x = DateTime()
        return x
    else:
        return str(x)

######################### TinyTable Class #########################

class TinyTable(
    OFS.SimpleItem.Item,
    Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager,
    ):
    """TinyTable is a product designed to manage a small amount of
tabular data.  It's intended to fill the gap between a Tabula or an Z
SQL Methods accessed SQL table, which are overkill for many tasks, and
folder token properties, which allow only a single "column".  TinyTable
also makes it possible to look up an item within the list, or to return
a subset of the list rows where columns equal particular values. 

TinyTable Properties

  Delimiter

    The *delimiter* is a single character that is used to
    separate the columns in the data input window. It must be a
    character that will not itself appear in the data. The default
    is ':'. 

  Columns

    *Columns* is a list of one or more column names separated by
    spaces. Columns are string-typed by default, but may optionally
    be integers, long integers, floating-point, or DateTime if the
    column name is suffixed with ':int', ':long', ':float', or ':date'
    respectively. 

    The first column is special. An index will be built on this
    column for "lookup" use (see below). The index built is unique. 
    That is, if there are multiple rows with the same first-column
    value, only one row will appear in the index, and only one row
    will be returned from an index query. If this is a problem, use
    a filter on the first row instead (see below). 

  Data

    The data consists of newline-separated rows containing columns
    separated by the delimiter character. Any input data will be
    adjusted to conform to the column specification. If the row
    contains too many columns the excess will be trimmed. If the
    row contains to few columns, columns containing an empty string
    or 0 (as appropriate to the column type) will be added. String
    values in a column specified to take a number will be replaced
    by 0. 

  Querying a TinyTable

    Assume you have a table named MyTable. It has these properties:

    Columns::

      last first middle n:int x:long

    Delimiter::

      :

    and the following data::

      smith,john,x,0,0L
      smith,bob,x,0,0L
      smith,bob,z,0,0L
      jones,bob,y,0,0L
      jones,john,y,0,0L
      jones,john,z,0,0L

    The data can be queried from DTML in several ways:

      Full Query::

        <!--#in MyTable-->

        Iterates through all rows of the TinyTable. Within the
        region contained by 'in' tag, the column names will be
        available as variables and so can be insterted. For
        example on the first iteration, '<!--#var first-->' will
        be replaced with 'john'. 

      Index Query::

        <!--#in "MyTable('jones')"-->

        The passed argument will be looked up in the table's
        index of the first column. Because the index is unique,
        either zero (if no matching rows) or one (if any
        matching rows) rows will be iterated through. In this
        case, any *one* of the three rows with a last name of
        'jones' could be returned. The choice of which row is
        returned when multiple rows have the same index value is
        unspecified. 

      Filter Query::

        <!--#in "MyTable(last='jones')"-->
        <!--#in "MyTable(first='john')"-->
        <!--#in "MyTable(last='jones', middle='y')"-->

        When one or more named arguments is given, a filter
        query is performed. Each argument name must be the name
        of a column, and the corresponding value is compared
        against that column in each row. Only matching rows are
        returned. The first example above, in contrast with the
        index query example, returns *all three* rows where the
        last name is 'jones'. 

        While an Index Query operates only on the first column,
        a filter query can operate on any column. In the second
        exmple above, all three rows with the first name 'john'
        are returned. 

        Finally, multiple filters may be specified. In this
        case only rows matching all contraints are iterated
        through. In the third example above, only the two rows
        where the last name is 'jones' and the middle initial is
        'y' will be returned. 
        """

    # Specify a name for the item type:
    meta_type = 'TinyTable'

    # Specify a relative URL for the icon used to display icons:
    icon = 'misc_/TinyTable/icon'

    # Specify definitions for tabs:
    manage_options=(
	{"label":"Properties",  "action":"manage_main"},
	{"label":"Advanced",    "action":"manage_advancedForm"},
        {"label":"View",        "action":"manage_view"},
	{"label":"Security",    "action":"manage_access"},
	{"label":"About",       "action":"manage_about"},
	)

    # Specify how individual operations add up to "permissions":
    __ac_permissions__=(
	('View management screens', ('manage_tabs','manage_main',
	                             'manage_about','manage_advancedForm')),
	('Change permissions',      ('manage_access',)           ),
	('Change TinyTable',        ('manage_edit','manage_editData',
	                             'manage_advanced')),
	('Query TinyTable Data',    ('','index_html','manage_view')),
	)

    def __init__(self, id, title='', delim_char='', columns=''):
	self.id = id
        self._dataver = 0
	self._SetState(title, delim_char, columns)

        self._rows = []
        self._index = {}
        self._n_rows = 0

        self.class_name_ = self.class_file_ = ""
        self._v_brain = None

    # Provide a "View" interface:
    manage_view = HTMLFile("View", globals())

    # Provide a "About..." interface:
    manage_about = HTMLFile("About", globals())

    # Provide interface for changing properties:
    manage_main=HTMLFile('Edit', globals())
    def manage_edit(self, title, delim_char, columns, REQUEST=None):
	"""Change item properties

	Note that we return people to our own interface, not to
	the folder we were in before.
	"""
        self._SetState(title, delim_char, columns)

        # make existing data conform to new column specification
        self._rows = map(self._FixRow, self._rows)

        # and regenerte the index, incase the above changed any data
        self._GenerateIndex()

        return self.manage_editedDialog(REQUEST)

    def _SetState(self, title, delim_char, columns):
	self.title = title

        try:
            self.delim_char_ = delim_char[0]
        except:
            self.delim_char_ = ':'

        self._DigestColumns(columns)
        self._dataver = 0

    manage_advancedForm = HTMLFile("Advanced", globals())
    def manage_advanced(self, class_name, class_file, REQUEST=None):
        """Change Advanced Settings"""
        self.class_name_, self.class_file_ = class_name, class_file
        self._v_brain = getBrain(self.class_file_, self.class_name_, 1)
        return self.manage_editedDialog(REQUEST)

    def manage_editData(self, data, REQUEST=None):
	"""Change item data"""
        self._DigestData(data)

        return self.manage_editedDialog(REQUEST)

    def _DigestColumns(self, column_list):
        self._col_index = {}
        self._col_names = []
        self._types = []
        self._items = []
        cols = string.split(column_list)

        for col in cols:
            item = {}
            x = string.split(col, ':')
            self._col_names.append(x[0])
            item['name'] = x[0]
            t = StringValued
            if len(x) > 1:
                if x[1] == 'int':
                    t = IntValued
                elif x[1] == 'long':
                    t = LongValued
                elif x[1] == 'float':
                    t = FloatValued
                elif x[1] == 'date':
                    t = DateValued
            self._types.append(t)
            item['type'] = TypeCode(t)
            self._items.append(item)

        self.n_cols = len(self._col_names)
        self.index_column = self._col_names[0]

        col_num = 0
        for col in self._col_names:
            self._col_index[col] = col_num
            col_num = col_num + 1

    def _FixRow(self, row):
        # force row to match specified number of tables
        if len(row) > self.n_cols:
            row = row[:self.n_cols]
        elif len(row) < self.n_cols:
            row = row + (self.n_cols - len(row)) * ['']

        newrow = []
        for i in range(0, self.n_cols):
            newrow.append(CoerceType(row[i], self._types[i]))

        return newrow

    def _GenerateIndex(self):
        self._index = {}

        for i in range(0, self._n_rows):
            self._index[self._rows[i][0]] = i

    def _DigestData(self, data):
        self._rows = []
        self._n_rows = 0

        for line in string.split(data, '\n'):
            # Ignore blank lines
            if not len(line):
                continue

            row = string.split(line, self.delim_char_)
            row = map(string.strip, row)
            row = self._FixRow(row)

            self._rows.append(row)
            self._n_rows = self._n_rows + 1

            self._GenerateIndex()

    def edit_data_(self, what):
        if what == "data": 
            lines = []

            for row in self._rows:
                lines.append(string.join(map(str, row), self.delim_char_))

            return string.join(lines, '\n')
        elif what  == "cols":
            l = []

            for i in range(0, self.n_cols):
                l.append(self._col_names[i] + TypeNames[self._types[i]])

            return string.join(l, ' ')

    def index_html(self):
        """Returns an HTML representation of the TinyTable's data"""
        
        s = "<table border=1><tr><th>"
        s = s + string.join(self._col_names, "</th>\n<th>") + "</th></tr>\n"
        for row in self._rows:
            s = s + "<tr><td>" + \
                string.join(map(str, row), "</td>\n<td>") + "</td></tr>\n"
        return s + "</table>"

    def __call__(self, *args, **kargs):
        if hasattr(self, '_v_brain'):
            brain = self._v_brain
        else:
            brain = self._v_brain = getBrain(self.class_file_, self.class_name_)
                           
        if len(args) == 1:
            if self._index.has_key(args[0]):
                return Results(
                    (self._items, [self._rows[self._index[args[0]]]]),
                    brains=brain, parent=None)
            else:
                return None
        elif len(kargs):
            rf = RowFilter(self, kargs)
            l = []

            for i in range(0, self._n_rows):
                if rf(self._rows[i]):
                    l.append(self._rows[i])
            return Results((self._items, l), brains=brain, parent=None)
        else:
            return Results((self._items, self._rows),
                brains=brain, parent=None)

######################### Helper Classes #########################

class RowFilter:
    def __init__(self, table, rules):
        self.cols = []
        self.vals = []
        self.n_rules = 0

        for col in rules.keys():
            if table._col_index.has_key(col):
                self.cols.append(table._col_index[col])
                self.vals.append(rules[col])
                self.n_rules = self.n_rules + 1

    def __call__(self, row):
        for i in range(0, self.n_rules):
            if row[self.cols[i]] != self.vals[i]:
                return 0
        return 1
